//
class C_Function
{
  //============================================================================
  public: double CalcFunc (double &args [], //function arguments
                           int     amount)  //amount of runs functions
  {
    double x, y;
    double sum = 0.0;
    for (int i = 0; i < amount; i++)
    {
      x = args [i * 2];
      y = args [i * 2 + 1];
      
      if (x < GetMinX ()) x = GetMinX ();
      if (x > GetMaxX ()) x = GetMaxX ();
      
      if (y < GetMinY ()) y = GetMinY ();
      if (y > GetMaxY ()) y = GetMaxY ();

      sum += Core (x, y);
    }

    sum /= (double)amount;
    
    //if (MathClassify(sum) == -FP_NAN) Print ("F: ", sum, " -| ", amount);
    //if (MathClassify(sum) ==  FP_NAN) Print ("F: ", sum, " +| ", amount);

    return sum;
  }

  public: double GetMinX   () {return xMin;}
  public: double GetMaxX   () {return xMax;}

  public: double GetMinY   () {return yMin;}
  public: double GetMaxY   () {return yMax;}

  public: double GetMaxFuncX () {return xMaxFunc;}
  public: double GetMaxFuncY () {return yMaxFunc;}

  public: double GetMinFuncX () {return xMinFunc;}
  public: double GetMinFuncY () {return yMinFunc;}

  public: double GetMinFun () {return minFun;}
  public: double GetMaxFun () {return maxFun;}

  public: string GetNamFun () {return fuName;}

  protected: //==================================================================
  protected: void SetMinX     (double min) {xMin = min;}
  protected: void SetMaxX     (double max) {xMax = max;}

  protected: void SetMinY     (double min) {yMin = min;}
  protected: void SetMaxY     (double max) {yMax = max;}

  protected: void SetMaxFuncX (double max) {xMaxFunc = max;}
  protected: void SetMaxFuncY (double max) {yMaxFunc = max;}

  protected: void SetMinFuncX (double max) {xMinFunc = max;}
  protected: void SetMinFuncY (double max) {yMinFunc = max;}

  protected: void SetMinFun   (double min) {minFun = min;}
  protected: void SetMaxFun   (double max) {maxFun = max;}

  protected: void SetNamFun (string nam) {fuName = nam;}

  //============================================================================
  private: virtual double Core (double x, double y) { return 0.0;}
  private: double xMin;
  private: double xMax;
  private: double yMin;
  private: double yMax;

  private: double xMaxFunc;
  private: double yMaxFunc;
  private: double xMinFunc;
  private: double yMinFunc;

  private: double minFun;
  private: double maxFun;
  private: string fuName;
};
//

//
enum EFunc
{
  Skin,
  Forest,
  Megacity,
  Rastrigin,
  Universe
};
C_Function *SelectFunction (EFunc f)
{
  C_Function *func;
  switch (f)
  {
    case  Skin:
      func = new C_Skin (); return (GetPointer (func));
    case  Forest:
      func = new C_Forest (); return (GetPointer (func));
    case  Megacity:
      func = new C_Megacity (); return (GetPointer (func));
    case  Rastrigin:
      func = new C_Rastrigin (); return (GetPointer (func));
    case  Universe:
      func = new C_Universe (); return (GetPointer (func));
    
    default:
      func = new C_Skin (); return (GetPointer (func));
  }
}
//

//
class C_Skin : public C_Function
{
  public: //===================================================================
  C_Skin ()
  {
    SetNamFun ("Skin");

    SetMinX (1.1950745);
    SetMaxX (2.39507448);
    SetMinY (2.1659335);
    SetMaxY (4.4659335);

    SetMaxFun   (4.918381646412932);  //1 points
    SetMaxFuncX (1.2485549205428181);
    SetMaxFuncY (4.345068261538417);

    SetMinFun   (-4.1808);            //1 points
    SetMinFuncX (1.7950745);
    SetMinFuncY (3.3159335);
  }

  private: //===================================================================
  double Core (double x, double y)
  {
    double a1=2*x*x;
    double a2=2*y*y;
    double b1=MathCos(a1)-1.1;
    b1=b1*b1;
    double c1=MathSin(0.5*x)-1.2;
    c1=c1*c1;
    double d1=MathCos(a2)-1.1;
    d1=d1*d1;
    double e1=MathSin(0.5*y)-1.2;
    e1=e1*e1;

   double res=b1+c1-d1+e1;
   return(res);
  }
};
//

//
class C_Forest : public C_Function
{
  public: //===================================================================
  C_Forest ()
  {
    SetNamFun ("Forest");

    SetMinX (-43.5);
    SetMaxX (-39.0);
    SetMinY (-47.35);
    SetMaxY (-40.0);

    SetMaxFun   (1.7678766100234535); //1 points
    SetMaxFuncX (-40.840704496667314);
    SetMaxFuncY (-41.982297150257104);

    SetMinFun   (0.0);           //many points
    SetMinFuncX (1.7950745);
    SetMinFuncY (3.3159335);
  }

  private: //===================================================================
  double Core (double x, double y)
  {
    double a = MathSin (MathSqrt (MathAbs (x - 1.13) + MathAbs (y - 2.0)));
    double b = MathCos (MathSqrt (MathAbs (MathSin (x))) + MathSqrt (MathAbs (MathSin (y - 2.0))));
    double f = a + b;

    double res = MathPow (f, 4);
    if (res < 0.0) res = 0.0;
    return (res);
  }
};
//

//
class C_Megacity : public C_Function
{
  public: //===================================================================
  C_Megacity ()
  {
    SetNamFun ("Megacity");

    SetMinX (-10.0);
    SetMaxX (-2.0);
    SetMinY (-10.5);
    SetMaxY (10);

    SetMaxFun   (12.0);          //1 points
    SetMaxFuncX (-3.12);
    SetMaxFuncY (1.95);

    SetMinFun   (0.0);           //many points
    SetMinFuncX (-6.0);
    SetMinFuncY (0.0);
  }

  private: //===================================================================
  double Core (double x, double y)
  {
    double a = MathSin (MathSqrt (MathAbs (x - 1.13) + MathAbs (y - 2.0)));
    double b = MathCos (MathSqrt (MathAbs (MathSin (x))) + MathSqrt (MathAbs (MathSin (y - 2.0))));
    double f = a + b;

    double res = floor (MathPow (f, 4));
    return (res);
  }
};
//

//
class C_Rastrigin : public C_Function
{
  public: //===================================================================
  C_Rastrigin ()
  {
    SetNamFun ("Rastrigin");
  
    SetMinX (-5.0);
    SetMaxX ( 5.0);
    SetMinY (-5.0);
    SetMaxY ( 5.0);
    
    SetMaxFun   (80.70658038767793);    //4 points
    SetMaxFuncX (4.522993660076278);
    SetMaxFuncY (4.522993659172958);

    SetMinFun   (0.0);                  //1 points
    SetMinFuncX (0.0);
    SetMinFuncY (0.0);
  }

  private: //===================================================================
  double Core (double x, double y)
  {
    double res = 20.0 + x * x + y * y - 10.0 * cos (2.0 * M_PI * x) - 10.0 * cos (2.0 * M_PI * y);
    return (res);
  }
};
//

//
class C_Universe : public C_Function
{
  public: //===================================================================
  C_Universe ()
  {
    SetNamFun ("Universe");

    SetMinX (-100.0);
    SetMaxX ( 100.0);
    SetMinY (-100.0);
    SetMaxY ( 100.0);

    SetMaxFun   (0.0); //1 points
    SetMaxFuncX (0.0);
    SetMaxFuncY (0.0);

    SetMinFun   (0.0);           //many points
    SetMinFuncX (0.0);
    SetMinFuncY (0.0);
  }

  private: //===================================================================
  double Core (double x, double y)
  {
    return (0.0);
  }
};
//